# -*- coding: utf-8 -*-

import os
import requests
from requests.structures import CaseInsensitiveDict

from .common import *
from .portable import utf8_string
from .error import RequestError
from .progress import LimitedReader, file_object_left_bytes
from .mime import mime_type


_CHUNK_SIZE = 8 * 1024
  

class Session(object):
    def __init__(self):
        self.session = requests.Session()

        pool_size = common.connect_pool_size
        self.session.mount('http://', requests.adapters.HTTPAdapter(pool_connections=pool_size, pool_maxsize=pool_size))

    def do_request(self, req, timeout):
        try:
            return Response(self.session.request(
                req.method,
                req.url,
                data=req.data,
                params=req.params,
                headers=req.headers,
                stream=True,
                timeout=timeout))
        except requests.RequestException as e:
            raise RequestError(e)


class Request(object):
    def __init__(self, method, url, data=None, params=None, headers=None):
        self.method = method
        self.url = url
        self.data = _to_http_body(data)
        self.params = params or {}

        if not isinstance(headers, CaseInsensitiveDict):
            self.headers = CaseInsensitiveDict(headers)
        else:
            self.headers = headers

        # ...
        if 'Accept-Encoding' not in self.headers:
            self.headers['Accept-Encoding'] = None


class Response(object):
    def __init__(self, response):
        self.response = response
        self.status = response.status_code
        self.headers = response.headers

        # When a response contains no body, iter_content() cannot
        # be run twice (requests.exceptions.StreamConsumedError will be raised).
        # For details of the issue, please see issue #82
        #
        # To work around this issue, we simply return b'' when everything has been read.
        #
        # Note you cannot use self.response.raw.read() to implement self.read(), because
        # raw.read() does not uncompress response body when the encoding is gzip etc., and
        # we try to avoid depends on details of self.response.raw.
        self.__all_read = False
      
    def read(self, amt=None):
        if self.__all_read:
            return b''
        
        if amt is None:
            content_list = []
            for chunk in self.response.iter_content(_CHUNK_SIZE):
                content_list.append(chunk)
            content = b''.join(content_list)
        
            self.__all_read = True
            return content
        
        try:
            return next(self.response.iter_content(amt))
        except StopIteration:
            self.__all_read = True
            return b''
    
    def __iter__(self):
        return self.response.iter_content(_CHUNK_SIZE)


def _to_http_body(data):
    data = utf8_string(data)

    if hasattr(data, '__len__'):
        return data
    
    if hasattr(data, 'seek') and hasattr(data, 'tell'):
        return LimitedReader(data, file_object_left_bytes(data))
    
    return data


def set_content_type(headers, name):
    """根据文件名在headers里设置Content-Type。如果headers中已经存在Content-Type，则直接返回。"""
    headers = headers or {}

    if 'Content-Type' in headers:
        return headers

    ext = os.path.splitext(name)
    ext_name = 'default'
    if len(ext) == 2:
        ext_name = ext[1][1:]

    content_type = mime_type(ext_name)
    if content_type:
        headers['Content-Type'] = content_type

    return headers




# -*- coding: utf8 -*-


import os

from .error import ClientError
from .portable import to_bytes


_CHUNK_SIZE = 8 * 1024


class LimitedReader(object):
    """限制读， 可以限定读取的长度 """
    def __init__(self, file_object, size):
        self.file_object = file_object
        self.size = size
        self.offset = 0
    
    def read(self, size=None):
        if self.offset >= self.size:
            return ''
        
        if (size is None or size < 0) or (size + self.offset >= self.size):
            data = self.file_object.read(self.size - self.offset)
            self.offset = self.size
            return data
        
        self.offset += size
        return self.file_object.read(size)

    @property
    def len(self):
        return self.size


def file_object_left_bytes(file_object):
    now_pos = file_object.tell()

    file_object.seek(0, os.SEEK_END)
    end = file_object.tell()
    file_object.seek(now_pos, os.SEEK_SET)

    return end - now_pos


def _get_data_length(data):
    if hasattr(data, '__len__'):
        return len(data)
    
    if hasattr(data, 'len'):
        return data.len
    
    if hasattr(data, 'seek') and hasattr(data, 'tell'):
        return file_object_left_bytes(data)

    return None


def _invoke_progress(progress_callback, transfered_bytes, total_bytes):
    if progress_callback:
        progress_callback(transfered_bytes, total_bytes)


class _IterableAdapter(object):
    def __init__(self, data, progress_callback=None):
        self.iter = iter(data)
        self.progress = progress_callback
        self.offset = 0

    def __iter__(self):
        return self
      
    def __next__(self):
        return self.next()

    def next(self):
        _invoke_progress(self.progress, self.offset, None)

        content = next(self.iter)
        self.offset += len(content)

        return content


class _ReadableAdapter(object):
    def __init__(self, file_object, progress_callback=None, chunk_size=None):
        self.file_object = file_object
        self.progress = progress_callback
        self.chunk_size = chunk_size
        self.offset = 0

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()
    
    def next(self):
        content = self.read(self.chunk_size)

        if content:
            return content

        return StopIteration

    def read(self, size=None):
        content = self.file_object.read(size)
        _invoke_progress(self.progress, self.offset, None)

        if content:
            self.offset += len(content)

        return content

  
class _BufferAdapter(object):
    def __init__(self, data, progress_callback=None, chunk_size=None, size=None):
        self.data = to_bytes(data)
        self.progress = progress_callback
        self.chunk_size = chunk_size
        self.size = size
        self.offset = 0

    @property
    def len(self):
        return self.size

    # for python 2.x
    def __bool__(self):
        return True
    # for python 3.x
    __nonzero__=__bool__

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        content = self.read(self.chunk_size)
        if content:
            return content
        
        return StopIteration

    def read(self, size=None):
        if self.offset >= self.size:
            return ''
        
        if size is None or size < 0:
            bytes_to_read = self.size - self.offset
        else:
            bytes_to_read = min(size, self.size - self.offset)
        
        if isinstance(self.data, bytes):
            content = self.data[self.offset:self.offset+bytes_to_read]
        else:
            content = self.data.read(bytes_to_read)
        
        self.offset += bytes_to_read
        _invoke_progress(self.progress, min(self.offset, self.size), self.size)

        return content


def make_progress_adapter(data, progress_callback, chunk_size, size=None):
    """返回一个适配器，从而在读取 `data` ，即调用read或者对其进行迭代的时候，能够
      调用进度回调函数。当 `size` 没有指定，且无法确定时，上传回调函数返回的总字节数为None。

    :param data: 可以是bytes、file object或iterable
    :param progress_callback: 进度回调函数，参见 :ref:`progress_callback`
    :param size: 指定 `data` 的大小，可选

    :return: 能够调用进度回调函数的适配器
    """
    data = to_bytes(data)

    if size is None:
        size = _get_data_length(data)

    if size is None:
        if hasattr(data, 'read'):
            return _ReadableAdapter(data, progress_callback, chunk_size)
        elif hasattr(data, '__iter__'):
            return _IterableAdapter(data, progress_callback)
        else:
            raise ClientError('{0} is not a file object, nor an iterator'.format(data.__class__.__name__))
    else:
        return _BufferAdapter(data, progress_callback, chunk_size, size)




# -*- coding: utf8 -*-

from .misc import *


class Result(object):
    def __init__(self, resp):
        self.resp = resp
        self.status = resp.status
        self.headers = resp.headers
        self.request_id = resp.headers.get('x-amz-request-id', '')

  
# etag value is '"etag"' so need strip the `"`.
def _get_etag(headers):
    return misc._get_dict_value(headers, 'etag', lambda x: x.strip('"'))


class PutObject(Result):
    def __init__(self, resp):
        super(PutObject, self).__init__(resp)

        self.etag = _get_etag(self.headers)


class InitMultipartUpload(Result):
    def __init__(self, resp):
        super(InitMultipartUpload, self).__init__(resp)

        self.upload_id = None


# 列出bucket下面的所有对象。
class ListObjects(Result):
    def __init__(self, resp):
        super(ListObjects, self).__init__(resp)
    
        self.is_continous = False
        self.next_marker = ''
        self.objects = []
        self.prefixs = []
  

class ListBuckets(Result):
    def __init__(self, resp):
        super(ListBuckets, self).__init__(resp)
        self.is_continuous = False
        self.next_marker = ''
        self.buckets = []


class ListMultipartUploads(Result):
    def __init__(self, resp):
        super(ListMultipartUploads, self).__init__(resp)
        self.is_continous = False
        self.next_key_marker = ''
        self.next_upload_id_marker = ''
        self.uploads = []
        self.prefixs = []


class ListParts(Result):
    def __init__(self, resp):
        super(ListParts, self).__init__(resp)

        self.is_continous = False
        self.next_marker = ''
        self.parts = []




# -*- coding: utf8 -*-

import socket


_ALPHA_AND_DIGIT = 'abcdefghijklmnopqrstuvwxyz0123456789'
_CONNECTOR = '-'
_BUCKET_CHAR_SET = set(_ALPHA_AND_DIGIT + _CONNECTOR) 


def _get_dict_value(d, key, converter=lambda x: x):
    if key in d:
        return converter(d[key])
    return None


def is_ipaddr(netloc):
    loc = netloc.split(':')
    try:
        socket.inet_aton(str(loc))
    except socket.error:
        return False
    
    return True


def is_localhost(netloc):
    loc = netloc.split(':')[0]
    return loc == 'localhost'


def bucket_name_is_valid(name):
    if len(name) < 3 or len(name) > 63:
        return False

    if name[-1] == _CONNECTOR:
        return False
    
    if name[0] not in _ALPHA_AND_DIGIT:
        return False
    
    return set(name) <= _BUCKET_CHAR_SET




# -*- coding: utf8 -*-

import time

from .client import * 
from .error import *
from .misc import *
from .progress import * 
from .url import * 
from .xml_body import *
from .limited import * 
from .common import get_value, connect_timeout, get_logger
from .portable import to_string, unicode_string
from .result import PutObject, InitMultipartUpload, ListBuckets, ListObjects


def _normalize_endpoint(endpoint):
    if endpoint.startswith('http://') or endpoint.startswith('https://'):
        return endpoint
    return 'http://' + endpoint


class _Base(object):
    def __init__(self, auth, endpoint, is_cname, session, conn_timeout):
        self.auth = auth
        self.endpoint = _normalize_endpoint(endpoint.strip())
        self.session = session or client.Session()
        self.timeout = get_value(conn_timeout, connect_timeout)
        
        self._url_make = url._Maker(self.endpoint, is_cname)

    def _do(self, method, bucket, key, **kwargs):
        key = to_string(key)
        req = client.Request(method, self._url_make(bucket, key), **kwargs)

        get_logger().debug("Bucket Request url={0}".format(req.url))

        self.auth._sign_request(req, bucket, key)

        resp = self.session.do_request(req, timeout=self.timeout)
        if resp.status // 100 != 2:
            raise error.make_response_error(resp)

        content_length = misc._get_dict_value(resp.headers, 'content-length', int)
        if content_length is not None and content_length == 0:
            resp.read()
        
        return resp

    def _parse_result(self, resp, parse_func, cls):
        result = cls(resp)
        parse_func(result, resp.read())
        return result


class Service(_Base):
    """用于Service操作的类，如罗列用户所有的Bucket。

    用法 ::

      >>> import oss2
      >>> auth = oss2.Auth('your-access-key-id', 'your-access-key-secret')
      >>> service = oss2.Service(auth, 'oss-cn-hangzhou.aliyuncs.com')
      >>> service.list_buckets()
      <oss2.models.ListBucketsResult object at 0x0299FAB0>

    :param auth: 包含了用户认证信息的Auth对象
    :type auth: oss2.Auth

    :param str endpoint: 访问域名，如杭州区域的域名为oss-cn-hangzhou.aliyuncs.com

    :param session: 会话。如果是None表示新开会话，非None则复用传入的会话
    :type session: oss2.Session

    :param float connect_timeout: 连接超时时间，以秒为单位。
    :param str app_name: 应用名。该参数不为空，则在User Agent中加入其值。
        注意到，最终这个字符串是要作为HTTP Header的值传输的，所以必须要遵循HTTP标准。
    """
    def __init__(self, auth, endpoint, session=None, connect_timeout=None):
        super(Service, self).__init__(auth, endpoint, False, session, connect_timeout)

    def list_buckets(self, prefix='', marker='', max_keys=100):
        """根据前缀罗列用户的Bucket。
    
        :param str prefix: 只罗列Bucket名为该前缀的Bucket，空串表示罗列所有的Bucket
        :param str marker: 分页标志。首次调用传空串，后续使用返回值中的next_marker
        :param int max_keys: 每次调用最多返回的Bucket数目
    
        :return: 罗列的结果
        :rtype: oss2.models.ListBucketsResult
        """
        resp = self._do('GET', '', '',
                        params={'prefix': prefix,
                                'marker': marker,
                                'max-keys': str(max_keys)})
        return self._parse_result(resp, xml_body.parse_list_buckets, ListBuckets)


class Bucket(_Base):
    """用于Bucket和Object操作的类，诸如创建、删除Bucket，上传、下载Object等。

    用法 ::

      >>> import dawn
      >>> auth = dawn.Authorize('your-access-key-id', 'your-access-key-secret')
      >>> bucket = dawn.Bucket(auth, 'http://<domain>', 'your-bucket')
      >>> bucket.put_object('readme.txt', 'content of the object')
        <dawn.result.PutObject object at 0x029B9930>

    :param auth: 包含了用户认证信息的Authorize对象
    :type auth: dawn.Authorize

    :param str endpoint: 访问域名或者CNAME
    :param str bucket: Bucket名
    :param bool is_cname: 如果endpoint是CNAME则设为True；反之，则为False。

    :param session: 会话。如果是None表示新开会话，非None则复用传入的会话
    :type session: dawn.http.Session

    :param float connect_timeout: 连接超时时间，以秒为单位。
    """
    def __init__(self, auth, endpoint, bucket, is_cname=False, session=None, connect_timeout=None):
        super(Bucket, self).__init__(auth, endpoint, is_cname, session, connect_timeout)

        self.bucket = bucket.strip()

    def sign_url(self, key, expires, method='GET', headers=None, params=None):
        """生成签名URL。

        常见的用法是生成加签的URL以供授信用户下载，如为log.jpg生成一个5分钟后过期的下载链接::

          >>> bucket.sign_url('GET', 'log.jpg', 5 * 60)
          'http://<oss-domain>/<bucket>/logo.jpg?OSSAccessKeyId=YourAccessKeyId\&Expires=1447178011&Signature=UJfeJgvcypWq6Q%2Bm3IJcSHbvSak%3D'

        :param key: 文件名
        :param expires: 过期时间（单位：秒），链接在当前时间再过expires秒后过期
        :param method: HTTP方法，如'GET'、'PUT'、'DELETE'等, 默认为GET
        :type method: str

        :param headers: 需要签名的HTTP头部，如名称以x-amz-meta-开头的头部（作为用户自定义元数据）、
          Content-Type头部等。对于下载，不需要填。
        :type headers: 可以是dict，建议是http.CaseInsensitiveDict

        :param params: 需要签名的HTTP查询参数

        :return: 签名URL。
        """
        key = to_string(key)
        req = client.Request(method, self._url_make(self.bucket, key), headers=headers, params=params)

        return self.auth._sign_url(req, self.bucket, key, expires)

    def put_object(self, key, data, headers=None, progress_callback=None, upload_speed_limited=None):
        headers = client.set_content_type(client.CaseInsensitiveDict(headers), key)

        dlen = progress._get_data_length(data)
        chunk_size = progress._CHUNK_SIZE
        if upload_speed_limited:
            data = limited.DelayReader(data, upload_speed_limited)
            chunk_size = data.chunk_size
            get_logger().info('uplpad speed limit(bytes/s)={0}, chunk={1}'.format(upload_speed_limited, chunk_size))
            print('uplpad speed limit(bytes/s)={0}, chunk={1}'.format(upload_speed_limited, chunk_size))

        if progress_callback:
            data = progress.make_progress_adapter(data, progress_callback, chunk_size, dlen)

        begin = time.time()
        resp = self.__do_object('PUT', key, data=data, headers=headers)
        end = time.time()
        get_logger().info('put object speed(bytes/s)={0}'.format(int(dlen / (end - begin))))
        print('put object speed(bytes/s)={0}'.format(int(dlen / (end - begin))))
        return PutObject(resp)

    def put_object_from_file(self, key, filename, headers=None, progress_callback=None, upload_speed_limited=None):
        headers = client.set_content_type(client.CaseInsensitiveDict(headers), filename)

        with open(unicode_string(filename), 'rb') as f:
            return self.put_object(key, f, headers=headers, progress_callback=progress_callback, upload_speed_limited=upload_speed_limited)


    def init_multipart_upload(self, key, headers=None):
        """初始化分片上传。

        返回值中的 `upload_id` 以及Bucket名和Object名三元组唯一对应了此次分片上传事件。

        :param str key: 待上传的文件名

        :param headers: HTTP头部
        :type headers: 可以是dict，建议是http.CaseInsensitiveDict

        :return: :class:`InitMultipartUpload<dawn.result.InitMultipartUpload>`
        """
        headers = client.set_content_type(client.CaseInsensitiveDict(headers), key)

        resp = self.__do_object('POST', key, params={'uploads': ''}, headers=headers)

        parser = xml_body.Parser(resp)
        
        return parser.init_multipart_upload()

    def upload_part(self, key, upload_id, part_number, data, progress_callback=None, upload_speed_limited=None, headers=None):
        """上传一个分片。

        :param str key: 待上传文件名，这个文件名要和 :func:`init_multipart_upload` 的文件名一致。
        :param str upload_id: 分片上传ID
        :param int part_number: 分片号，最小值是1.
        :param data: 待上传数据。
        :param progress_callback: 用户指定进度回调函数。可以用来实现进度条等功能。参考 :ref:`progress_callback` 。

        :param headers: 用户指定的HTTP头部。可以指定Content-MD5头部等
        :type headers: 可以是dict，建议是http.CaseInsensitiveDict

        :return: :class:`PutObject <dawn.result.PutObject>`
        """
        dlen = progress._get_data_length(data)
        chunk_size = progress._CHUNK_SIZE
        if upload_speed_limited:
            data = limited.DelayReader(data, upload_speed_limited)
            chunk_size = data.chunk_size
            get_logger().info('uplpad speed limit(bytes/s)={0}, chunk={1}'.format(upload_speed_limited, chunk_size))
            print('uplpad speed limit(bytes/s)={0}, chunk={1}'.format(upload_speed_limited, chunk_size))

        if progress_callback:
            data = progress.make_progress_adapter(data, progress_callback, chunk_size, dlen)

        begin = time.time()
        resp = self.__do_object('PUT', key,
                                params={'uploadId': upload_id, 'partNumber': str(part_number)},
                                headers=headers,
                                data=data)
        end = time.time()
        get_logger().info('upload multipart speed(bytes/s)={0}'.format(int(dlen / (end - begin))))
        print('upload multipart speed(bytes/s)={0}'.format(int(dlen / (end - begin))))
        return PutObject(resp)

    def complete_multipart_upload(self, key, upload_id, parts, headers=None):
        """完成分片上传，创建文件。

        :param str key: 待上传的文件名，这个文件名要和 :func:`init_multipart_upload` 的文件名一致。
        :param str upload_id: 分片上传ID

        :param parts: PartInfo列表。PartInfo中的part_number和etag是必填项。其中的etag可以从 :func:`upload_part` 的返回值中得到。
        :type parts: list of `PartInfo <dawn.xml_body.PartInfo>`

        :param headers: HTTP头部
        :type headers: 可以是dict，建议是http.CaseInsensitiveDict

        :return: :class:`PutObject <dawn.result.PutObject>`
        """
        maker = xml_body.Maker()

        data = maker.complete_upload_request(sorted(parts, key=lambda p: p.part_number))
        resp = self.__do_object('POST', key,
                                params={'uploadId': upload_id},
                                data=data,
                                headers=headers)

        return PutObject(resp)

    def abort_multipart_upload(self, key, upload_id):
        """取消分片上传。

        :param str key: 待上传的文件名，这个文件名要和 :func:`init_multipart_upload` 的文件名一致。
        :param str upload_id: 分片上传ID

        :return: :class:`Result <dawn.result.Result>`
        """
        resp = self.__do_object('DELETE', key,
                                params={'uploadId': upload_id})
        return Result(resp)

    def list_objects(self, prefix='', delimiter='', marker='', max_keys=100):
        """根据前缀罗列Bucket里的文件。
 
        :param str prefix: 只罗列文件名为该前缀的文件
        :param str delimiter: 分隔符。可以用来模拟目录
        :param str marker: 分页标志。首次调用传空串，后续使用返回值的next_marker
        :param int max_keys: 最多返回文件的个数，文件和目录的和不能超过该值
 
        :return: :class:`ListObjectsResult <oss2.models.ListObjectsResult>`
        """
        resp = self.__do_object('GET', '',
                                params={'prefix': prefix,
                                        'delimiter': delimiter,
                                        'marker': marker,
                                        'max-keys': str(max_keys),
                                        'encoding-type': 'url'})
        return self._parse_result(resp, xml_body.parse_list_objects, ListObjects)

    def __do_object(self, method, key, **kwargs):
        return self._do(method, self.bucket, key, **kwargs)

    def __do_bucket(self, method, **kwargs):
        return self._do(method, self.bucket_name, '', **kwargs)




# -*- coding: utf8 -*-

import math
import time

from monotonic import monotonic
from error import ClientError


class DelayReader(object):
    def __init__(self, data, bytes_per_second, times_per_second=20, sleep_func=time.sleep):
        self._data = data
        self._bytes_per_second = bytes_per_second
        self._times_per_second = times_per_second
        self._sleep = sleep_func
        self._per_elapsed = 1.0 / times_per_second
        if self._bytes_per_second < 1024:
          self._bytes_per_second = 1024
        self._chunk = self._bytes_per_second // times_per_second
        self._now = None
        self._next_time = None
        self._bucket = self._chunk
        #print(self._chunk)

    def __iter__(self):
        return self
    
    def __next__(self):
        return self.next()

    def next(self):
        content = self.read(self._chunk)

        if content:
            return content

        return StopIteration

    def _update_elapsed(self):
        now = monotonic()
        if now > self._next_time:
            elapsed = now - self._next_time
            ticks = int(math.ceil(elapsed / self._per_elapsed))
            self._bucket += ticks * self._chunk
            self._next_time += ticks * self._per_elapsed
    
    def _read(self, size=None):
        if hasattr(self._data, 'read'):
            return self._data.read(size)
        else:
            raise ClientError('{0} is not a file object, '.format(data.__class__.__name__))

    def read(self, size=None):
        if self._now is None:
            self._now = monotonic()
            self._next_time = self._now + self._per_elapsed
        
        if size is None:
            size = self._chunk

        self._update_elapsed()
        
        if size > self._bucket:
            now = monotonic()
            ticks = (size - self._bucket) / self._chunk
            time_df = self._next_time - now - self._per_elapsed
            delay = ticks * self._per_elapsed
            #print('ticks={0}, time_diff={1}, delay={2}'.format(ticks, time_df, delay))
            if time_df > 0:
                delay += time_df
            self._sleep(delay)
            self._update_elapsed()

        data = self._read(size)
        self._bucket -= size
        return data
    
    @property
    def chunk_size(self):
        return self._chunk


if __name__ == "__main__":
    with open('20m.vbox', 'rb') as f:
        dr = DelayReader(f, 1024000, 20)

        t_zero = time.time()
        while True:
            data = dr.read(2048000)
            print("Got %d at %0.2f" % (len(data),  time.time() - t_zero))
            if len(data) == 0:
                break
  



# -*- coding: utf8 -*-

# xml body for request or response.

import re

from lxml import etree

from .result import InitMultipartUpload
from .portable import unicode_string, to_string, urlunquote
from .common import iso8601_to_unixtime, SimplifiedBucketInfo, SimplifiedObjectInfo


utf8_parser = etree.XMLParser(encoding='utf-8')


def _find_tag(parent, path):
    child = parent.find(path, parent.nsmap)
    if child is None:
        raise RuntimeError("parse xml: " + path + " could not be found under " + parent.tag)

    if child.text is None:
        return ''

    return to_string(child.text)


def _find_object(parent, path, url_encoded):
    name = _find_tag(parent, path)
    if url_encoded:
        return urlunquote(name)
    else:
        return name


class Parser(object):
    def __init__(self, resp):
        self.resp = resp
        self.body = resp.read()
        self.root = etree.fromstring(self.body)
        self.xml_namespaces = self.root.nsmap
    
    def init_multipart_upload(self):
        result = InitMultipartUpload(self.resp)
        result.upload_id = to_string(self.root.find('UploadId', self.xml_namespaces).text)

        return result


def _add_text_child(parent, tag, text):
    etree.SubElement(parent, tag).text = unicode_string(text)


def _xml_to_string(root):
    return etree.tostring(root, encoding='utf-8')


class PartInfo(object):
    """表示分片信息的文件。

    该文件既用于 :func:`list_parts <dawn.Bucket.list_parts>` 的输出，也用于 :func:`complete_multipart_upload
    <dawn.Bucket.complete_multipart_upload>` 的输入。

    :param int part_number: 分片号
    :param str etag: 分片的ETag
    """
    def __init__(self, part_number, etag):
        self.part_number = part_number
        self.etag = etag


class Maker(object):
    def __init__(self):
        pass

    def complete_upload_request(self, parts):
        root = etree.Element('CompleteMultipartUpload')
        for p in parts:
            node = etree.SubElement(root, 'Part')
            _add_text_child(node, 'PartNumber', str(p.part_number))
            _add_text_child(node, 'ETag', '"{0}"'.format(p.etag))
        
        return _xml_to_string(root)


def parse_list_buckets(result, body):
    root = etree.fromstring(body)

    if root.find('IsTruncated') is None:
        result.is_truncated = False
    else:
        result.is_truncated = root.find('IsTruncated', root.nsmap)

    if result.is_truncated:
        result.next_marker = root.find('NextMarker', root.nsmap)

    for bucket_node in root.findall('Buckets/Bucket', root.nsmap):
        result.buckets.append(SimplifiedBucketInfo(
            name=bucket_node.find('Name', bucket_node.nsmap).text,
            creation_date=iso8601_to_unixtime(bucket_node.find('CreationDate', bucket_node.nsmap).text)
        ))


def _is_url_encoding(root):
    node = root.find('EncodingType', root.nsmap)
    if node is not None and to_string(node.text) == 'url':
        return True
    else:
        return False


def parse_list_objects(result, body):
    root = etree.fromstring(body)

    url_encoded = _is_url_encoding(root)

    for contents_node in root.findall('Contents', root.nsmap):
        result.objects.append(SimplifiedObjectInfo(
            _find_object(contents_node, 'Key', url_encoded),
            iso8601_to_unixtime(_find_tag(contents_node, 'LastModified')),
            _find_tag(contents_node, 'ETag').strip('"'),
            int(_find_tag(contents_node, 'Size')),
            _find_tag(contents_node, 'StorageClass')
        ))

    for prefix_node in root.findall('CommonPrefixes', root.nsmap):
        result.prefixes.append(_find_object(prefix_node, 'Prefix', url_encoded))

    return result




# -*- coding: utf8 -*-

_MIME_TYPE_DICT = {
    "html": "text/html",
    "htm": "text/html",
    "shtml": "text/html",
    "css": "text/css",
    "xml": "text/xml",
    "gif": "image/gif",
    "jpeg": "image/jpeg",
    "jpg": "image/jpeg",
    "js": "application/x-javascript",
    "atom": "application/atom+xml",
    "rss": "application/rss+xml",
    "mml": "text/mathml",
    "txt": "text/plain",
    "jad": "text/vnd.sun.j2me.app-descriptor",
    "wml": "text/vnd.wap.wml",
    "htc": "text/x-component",
    "png": "image/png",
    "tif": "image/tiff",
    "tiff": "image/tiff",
    "wbmp": "image/vnd.wap.wbmp",
    "ico": "image/x-icon",
    "jng": "image/x-jng",
    "bmp": "image/x-ms-bmp",
    "svg": "image/svg+xml",
    "svgz": "image/svg+xml",
    "webp": "image/webp",
    "jar": "application/java-archive",
    "war": "application/java-archive",
    "ear": "application/java-archive",
    "hqx": "application/mac-binhex40",
    "doc ": "application/msword",
    "pdf": "application/pdf",
    "ps": "application/postscript",
    "eps": "application/postscript",
    "ai": "application/postscript",
    "rtf": "application/rtf",
    "xls": "application/vnd.ms-excel",
    "ppt": "application/vnd.ms-powerpoint",
    "wmlc": "application/vnd.wap.wmlc",
    "kml": "application/vnd.google-earth.kml+xml",
    "kmz": "application/vnd.google-earth.kmz",
    "7z": "application/x-7z-compressed",
    "cco": "application/x-cocoa",
    "jardiff": "application/x-java-archive-diff",
    "jnlp": "application/x-java-jnlp-file",
    "run": "application/x-makeself",
    "pl": "application/x-perl",
    "pm": "application/x-perl",
    "prc": "application/x-pilot",
    "pdb": "application/x-pilot",
    "rar": "application/x-rar-compressed",
    "rpm": "application/x-redhat-package-manager",
    "sea": "application/x-sea",
    "swf": "application/x-shockwave-flash",
    "sit": "application/x-stuffit",
    "tcl": "application/x-tcl",
    "tk": "application/x-tcl",
    "der": "application/x-x509-ca-cert",
    "pem": "application/x-x509-ca-cert",
    "crt": "application/x-x509-ca-cert",
    "xpi": "application/x-xpinstall",
    "xhtml": "application/xhtml+xml",
    "zip": "application/zip",
    "wgz": "application/x-nokia-widget",
    "bin": "application/octet-stream",
    "exe": "application/octet-stream",
    "dll": "application/octet-stream",
    "deb": "application/octet-stream",
    "dmg": "application/octet-stream",
    "eot": "application/octet-stream",
    "iso": "application/octet-stream",
    "img": "application/octet-stream",
    "msi": "application/octet-stream",
    "msp": "application/octet-stream",
    "msm": "application/octet-stream",
    "mid": "audio/midi",
    "midi": "audio/midi",
    "kar": "audio/midi",
    "mp3": "audio/mpeg",
    "ogg": "audio/ogg",
    "m4a": "audio/x-m4a",
    "ra": "audio/x-realaudio",
    "3gpp": "video/3gpp",
    "3gp": "video/3gpp",
    "mp4": "video/mp4",
    "mpeg": "video/mpeg",
    "mpg": "video/mpeg",
    "mov": "video/quicktime",
    "webm": "video/webm",
    "flv": "video/x-flv",
    "m4v": "video/x-m4v",
    "mng": "video/x-mng",
    "asx": "video/x-ms-asf",
    "asf": "video/x-ms-asf",
    "wmv": "video/x-ms-wmv",
    "avi": "video/x-msvideo",
    "ts": "video/MP2T",
    "m3u8": "application/x-mpegURL",
    "apk": "application/vnd.android.package-archive",
    "default": "binary/octet-stream"
}

def mime_type(file_ext):
    file_ext = file_ext.lower()
    if file_ext in _MIME_TYPE_DICT.keys():
        return _MIME_TYPE_DICT[file_ext]
    
    return _MIME_TYPE_DICT['default']



# -*- coding: utf8 -*-

import re
import xml.etree.ElementTree as ElementTree
from xml.parsers import expat

from .portable import to_string


STATUS_ClIENT_ERROR = -1
STATUS_REQUEST_ERROR = -2


class BaseError(Exception):
    def __init__(self, status, headers, body, detail):
        # HTTP的状态码
        self.status = status
        # 请求id，用于跟踪一个请求。
        self.request_id = headers.get('x-amz-request-id', '')
        # HTTP响应体
        self.body = body
        # 详细的错误信息，通常是一个xml，但是经过解析后，被存在一个dict内。
        self.detail = detail

        # 错误信息中的错误码
        self.code = self.detail.get('Code', '')
        # 错误信息中的错误消息
        self.message = self.detail.get('Message', '')

    def __str__(self):
        error = {'HTTP-Code': self.status, 'request-id': self.request_id, 'detail': self.detail}
        return str(error)

    def _str_with_body(self):
        error = {'HTTP-Code': self.status, 'request-id': self.request_id, 'detail': self.body}
        return str(error)


class ClientError(BaseError):
    def __init__(self, message):
        BaseError.__init__(self, STATUS_CLIENT_ERROR, {}, 'ClientError: ' + message, {})
    
    def __str__(self):
        return self._str_with_body()
  

class RequestError(BaseError):
    def __init__(self, e):
        BaseError.__init__(self, STATUS_REQUEST_ERROR, {}, 'RequestError: ' + str(e), {})
        self.exception = e

    def __str__(self):
        return self._str_with_body()


class HTTPError(BaseError):
    pass


class NotFound(HTTPError):
    status = 404
    code = ''


class InvalidRequest(HTTPError):
    status = 400
    code = 'InvalidRequest'


class OperationNotSupported(HTTPError):
    status = 400
    code = 'OperationNotSupported'


class InvalidArgument(HTTPError):
    status = 400
    code = 'InvalidArgument'

    def __init__(self, status, headers, body, detail):
        super(InvalidArgument, self).__init__(status, headers, body, detail)
        self.name = detail.get('ArgumentName')
        self.value = detail.get('ArgumentValue')


class InvalidObjectName(HTTPError):
    status = 400
    code = 'InvalidObjectName'


class NoSuchBucket(NotFound):
    status = 404
    code = 'NoSuchBucket'


class NoSuchKey(NotFound):
    status = 404
    code = 'NoSuchKey'


class NoSuchUpload(NotFound):
    status = 404
    code = 'NoSuchUpload'


class AccessDenied(HTTPError):
    status = 403
    code = 'AccessDenied'


class SignatureDoesNotMatch(HTTPError):
    status = 403
    code = 'SignatureDoesNotMatch'


# 动态的获取子类和孙子类
def _iter_subclasses(cls):
    for sub in cls.__subclasses__():
        yield sub
        for sub_sub in _iter_subclasses(sub):
            yield sub_sub


def _parse_xml_body(body):
    try:
        root = ElementTree.fromstring(body)
        if root.tag != 'Error':
            return {}

        detail = {}
        for child in root:
            detail[child.tag] = child.text
        return detail
    except ElementTreeParseError:
        return _guess_xml_detail(body)


def _guess_xml_detail(body):
    detail = {}
    body = to_string(body)

    if '<Error>' not in body or '</Error>' not in body:
        return detail
    
    m = re.search('<Code>(.*)</Code>', body)
    if m:
        detail['Code'] = m.group(1)

    m = re.search('<Message>(.*)</Message>', body)
    if m:
        detail['Message'] = m.group(1)
    
    return detail


def make_response_error(resp):
    status = resp.status
    headers = resp.headers
    body = resp.read(4096)
    detail = _parse_xml_body(body)
    code = detail.get('Code', '')

    try:
        error = _STATUS_TO_ERROR_DICT[(status, code)]
        return error(status, headers, body, detail)
    except KeyError:
        return HTTPError(status, headers, body, detail)


# 生成 (status, code): 
_STATUS_TO_ERROR_DICT = {}
for cls in _iter_subclasses(HTTPError):
    status = getattr(cls, 'status', None)
    code = getattr(cls, 'code', None)

    if status is not None and code is not None:
        _STATUS_TO_ERROR_DICT[(status, code)] = cls

# XML parsing exceptions have changed in Python2.7 and ElementTree 1.3
if hasattr(ElementTree, 'ParseError'):
    ElementTreeParseError = (ElementTree.ParseError, expat.ExpatError)
else:
    ElementTreeParseError = (expat.ExpatError)




# -*- coding: utf-8 -*-

import hmac
import base64
import time
from hashlib import sha1 as sha
from email.utils import formatdate

from .url import *
from .portable import utf8_string, to_string, urlquote
from .common import get_logger


class Authorize(object):
    """用于保存用户的access_key/secret_key, 以及计算签名。"""
    def __init__(self, access_key, secret_key):
        self.access_key = access_key.strip()
        self.secret_key = secret_key.strip()

    def _sign_url(self, req, bucket, key, expires):
        future = int(time.time()) + expires

        req.headers['date'] = str(future)
        signature = self.__make_signature(req, bucket, key)

        req.params['AWSAccessKeyId'] = self.access_key
        req.params['Expires'] = str(future)
        req.params['Signature'] = signature

        return req.url + '?' + '&'.join(url._to_quoted_string(k, v) for k, v in req.params.items())

    def _sign_request(self, req, bucket, key):
        if 'date' not in req.headers and 'x-amz-date' not in req.headers:
            req.headers['date'] = formatdate(
              timeval=None,
              localtime=False,
              usegmt=True)

        signature = self.__make_signature(req, bucket, key)
        get_logger().debug('signature={0}'.format(signature))

        req.headers['Authorization'] = "AWS {0}:{1}".format(self.access_key, signature)    
        get_logger().debug('header[Authorization]={0}'.format(req.headers['Authorization']))

    def __make_signature(self, req, bucket, key):
        sign_string = self.__get_sign_string(req, bucket, key)

        key = utf8_string(self.secret_key)
        msg = utf8_string(sign_string)
          
        get_logger().debug('string to sign:\n{0}'.format(sign_string))

        h = hmac.new(key, msg, digestmod=sha)
        
        get_logger().debug('HMAC digest={0}'.format(h.digest()))
        return to_string(base64.b64encode(h.digest())).strip()
    
    def __get_sign_string(self, req, bucket, key):

        content_md5 = req.headers.get('content-md5', '')
        content_type = req.headers.get('content-type', '')
        date = ''
        if 'x-amz-date' not in req.headers:
            date = req.headers.get('date', '')

        canon_headers = self.__get_canon_headers(req)
        canon_resource = self.__get_canon_resource(req, bucket, key)

        return '\n'.join([req.method, content_md5, content_type, date, canon_headers + canon_resource])
    
    def __get_canon_headers(self, req):
        headers = req.headers
        canon_headers = []
        #todo: 头中相同的域需要合并。
        for k, v in headers.items():
            lk = k.lower()
            if lk.startswith('x-amz-'):
                canon_headers.append((lk, v))

        canon_headers.sort(key=lambda x: x[0])
        if canon_headers:
            return '\n'.join(k + ':' + v for k, v in canon_headers) + '\n'

        return ''

    def __get_canon_resource(self, req, bucket, key):
        if not bucket:
            return '/'
        key = utf8_string(key)
        key = urlquote(key, '')
        return '/{0}/{1}{2}'.format(bucket, key, self.__get_subresource(req.params))

    _subresource_keys = frozenset(
        [
            'acl', 'location', 'logging', 'partNumber', 'policy', 'requestPayment',
            'torrent', 'versioning', 'versionId', 'website', 'uploads', 'uploadId',
            'response-content-type', 'response-content-language', 'response-expires',
            'response-cache-control', 'delete', 'lifecycle', 'response-content-disposition',
            'response-content-encoding', 'tagging', 'notification', 'cors'
        ]
    )
    
    def __get_subresource(self, params):
        if not params:
            return ''
        
        subresource = []
        for k, v in params.items():
            if k in self._subresource_keys:
                subresource.append((k, v))
        
        subresource.sort(key=lambda k: k[0])

        if subresource:
            return '?' + '&'.join(self.__query_string(k, v) for k, v in subresource)
        
        return ''
    
    def __query_string(self, key, value):
        if value:
            return key + '=' + value
        
        return key




# -*- coding: utf8 -*-

import os
import config
import auth
import bucket
from xml_body import PartInfo
from progress import LimitedReader

def progress_func(transfered, total):
  pass
  #print('transfered:{0}, total:{1}'.format(transfered, total))

def prepare(access_key, secret_key, endpoint, is_cname, bucket_name):
  print('[prepare] ----> ')

  print('\taccess key={0}'.format(access_key))
  print('\tsecret key={0}'.format(secret_key))
  ath = auth.Authorize(access_key, secret_key)
  
  print('\tendpoint={0}'.format(endpoint))
  print('\tis CNAME={0}'.format(is_cname))

  print('\tbucket={0}'.format(bucket_name))
  
  return bucket.Bucket(ath, endpoint, bucket_name, is_cname)

def put_object_file(bkt, obj, file, limit=None):
  print('[put object from file] ----> ')

  print('\tobject={0}'.format(obj))
  print('\tfile={0}'.format(file))

  result = bkt.put_object_from_file(obj, file, progress_callback=progress_func, upload_speed_limited=limit)

  print('\tput object from file', result.status, result.request_id)
  return result

def put_object(bkt, obj, file, limit=None):
  print('[put object] ----> ')

  print('\tobject={0}'.format(obj))
  print('\tfile={0}'.format(file))

  with open(file, 'rb') as f:
    result = bkt.put_object(obj, f, progress_callback=progress_func, upload_speed_limited=limit)

  print('\tput object from file', result.status, result.request_id)
  return result

def generate_url(bkt, obj):
  print('[generate url] ----> ')
  sign_url = bkt.sign_url(obj, 60 * 60)
  print('\turl =', sign_url)

def multipart_upload(bkt, obj, file, limit=None):
  print('[multipart upload] ----> ')
  flen = os.path.getsize(file)
  with open(file, 'rb') as f:
    slice = (1024 * 1024) * 5

    times = flen // slice
    left = flen % slice

    print("\tfile://{0}'s length={1}, slice length={2}, times={3}, left={4}".format(file, flen, slice, times, left))

    result = bkt.init_multipart_upload(obj)
    uid = result.upload_id

    print('\tupload id={0}'.format(uid))

    parts = []
    number = 1
    while True:
      if times == 0:
        break
      
      reader = LimitedReader(f, slice)
      
      out = bkt.upload_part(obj, uid, number, reader, progress_callback=progress_func, upload_speed_limited=limit)
      part = PartInfo(number, out.etag)
      print('\tpart number={0}, ETag={1}'.format(number, out.etag))
      parts.append(part)

      number += 1
      times -= 1
    
    if left > 0:
      reader = LimitedReader(f, left)
      
      out = bkt.upload_part(obj, uid, number, reader, progress_callback=progress_func, upload_speed_limited=limit)
      part = PartInfo(number, out.etag)
      print('\tpart number={0}, ETag={1}'.format(number, out.etag))
      parts.append(part)

    result = bkt.complete_multipart_upload(obj, uid, parts)
  print('\tmultipart upload succeed...')


def list_bucket(access_key, secret_key, endpoint):
  authorization = auth.Authorize(access_key, secret_key)
  service = bucket.Service(authorization, endpoint)
  bucket_names = service.list_buckets(max_keys=10).buckets

  for bucket_name in bucket_names:
    print (bucket_name.name, bucket_name.creation_date)


def list_object(access_key, secret_key, endpoint, bucket_name):
  bkt = prepare(access_key, secret_key, endpoint, False, bucket_name)
  for item in bkt.list_objects().objects:
    print (item.key)


if __name__ == '__main__':
  conf = config.Config()

  access_key = conf.dict['access_key']
  secret_key = conf.dict['secret_key']
  endpoint = conf.dict['endpoint']
  is_cname = conf.dict['is_cname'] == 'true'
  bucket_name = conf.dict['bucket_name']

  bkt = prepare(access_key, secret_key, endpoint, is_cname, bucket_name)

  obj = conf.dict['object']
  file = conf.dict['file']
  try:
    limit = int(conf.dict['send_speed'])
  except:
    limit = 0
  
  print('\t speed limited={0}'.format(limit))
  put_object(bkt, obj, file)
  multipart_upload(bkt, obj, file, limit)

  generate_url(bkt, obj)

  list_bucket(access_key, secret_key, endpoint)

  list_object(access_key, secret_key, endpoint, "testing")



# -*- coding: utf8 -*-


from .misc import is_ipaddr, is_localhost
from .portable import urlquote, urlparse


class _Maker(object):
    def __init__(self, endpoint, is_cname):
        p = urlparse(endpoint)

        self.scheme = p.scheme
        self.netloc = p.netloc
        self.is_cname = is_cname

    def __call__(self, bucket, key):
        key = urlquote(key, '')

        # CNAME: 泛域名方式　http://<bucket>.<domain>/<key>
        if self.is_cname:
            return '{0}://{1}/{2}'.format(self.scheme, self.netloc, key)
        
        # ip格式或者localhost
        if is_ipaddr(self.netloc) or is_localhost(self.netloc):
            if bucket:
                return '{0}://{1}/{2}/{3}'.format(self.scheme, self.netloc, bucket, key)
            else:
                return '{0}://{1}/{2}'.format(self.scheme, self.netloc, key)

        # bucket和key都为空
        if not bucket:
            assert not key
            return '{0}://{1}'.format(self.scheme, self.netloc)

        # 我们自己的url http://<domain>/<bucket>/<key>
        return '{0}://{1}/{2}/{3}'.format(self.scheme, self.netloc, bucket, key)


def _to_quoted_string(k, v):
    if v:
        return urlquote(k, '') + '=' + urlquote(v, '')
    
    return urlquote(k, '')



# -*- coding: utf8 -*-

# config reader


class Config(object):
    def __init__(self, conf_file='config.txt'):
        self._dict = {}
        with open(conf_file, 'r') as f:
            for line in f:
                line = line.strip()
                if line and not line.startswith('//'):
                    p = line.split('=', 1)
                    if len(p) == 2:
                        self._dict[p[0]] = p[1]
    
    @property
    def dict(self):
        return self._dict
        



# -*- coding: utf-8 -*-

import logging
import datetime
import calendar
import re


# 连接超时，单位秒
connect_timeout = 30

connect_pool_size = 10

_ISO8601_RE = re.compile(
    r'(?P<year>\d+)-(?P<month>01|02|03|04|05|06|07|08|09|10|11|12)-(?P<day>0[1-9]|([1-2]\d)|(3[0-1]))T(?P<hour>([0-1]\d)|(2[0-3])):(?P<minute>[0-5]\d):(?P<second>[0-5]\d)\.\d\d\dZ$'
)


logger = logging.getLogger()


def get_value(value, default_value):
    if value is None:
        return default_value

    return value


def get_logger():
    return logger


def iso8601_to_unixtime(time_string):
    """把ISO8601时间字符串（形如，2012-02-24T06:07:48.000Z）转换为UNIX时间，精确到秒。"""

    m = _ISO8601_RE.match(time_string)

    if not m:
        raise ValueError(time_string + " is not in valid ISO8601 format")

    day = int(m.group('day'))
    month = int(m.group('month'))
    year = int(m.group('year'))
    hour = int(m.group('hour'))
    minute = int(m.group('minute'))
    second = int(m.group('second'))

    tm = datetime.datetime(year, month, day, hour, minute, second).timetuple()

    return calendar.timegm(tm)


class SimplifiedBucketInfo(object):
    """:func:`list_buckets <oss2.Service.list_objects>` 结果中的单个元素类型。"""
    def __init__(self, name, creation_date):
        #: Bucket名
        self.name = name

        #: Bucket的创建时间，类型为int。参考 :ref:`unix_time`。
        self.creation_date = creation_date


class SimplifiedObjectInfo(object):
    #def __init__(self, key, last_modified, etag, type, size, storage_class):
    def __init__(self, key, last_modified, etag, size, storage_class):
        #: 文件名，或公共前缀名。
        self.key = key

        #: 文件的最后修改时间
        self.last_modified = last_modified

        #: HTTP ETag
        self.etag = etag

        #: 文件类型
        # self.type = type

        #: 文件大小
        self.size = size

        #: 文件的存储类别，是一个字符串。
        self.storage_class = storage_class

    def is_prefix(self):
        """如果是公共前缀，返回True；是文件，则返回False"""
        return self.last_modified is None




#!/usr/bin/env python

import re

import setuptools

version = ""
with open('dawn_sdk/__init__.py', 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="dawn-sdk",
    version=version,
    author="TAL AI-LAB",
    author_email="author@example.com",
    description="This is the SDK for dawn, which is OSS system. Dawn is honorly presented by TAL's ocean team.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://dawn.shareurl.facethink.com",
    install_requires=[
        'requests!=2.9.0',
        'lxml',
        'monotonic',
    ],
    packages=setuptools.find_packages(exclude=("test")),
    classifiers=(
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5"
    ),
    exclude_package_data={'': ["dawn_sdk/test.py", "dawn_sdk/config.txt"]},
)

